from typing import Dict

from .groups import RoundRobin
from rstt.ranking.ranking import Ranking
from rstt import BetterWin
from rstt.stypes import Solver

import random

class RandomRound(RoundRobin):
    def __init__(self, name: str, seeding: Ranking, solver: Solver=BetterWin(), cashprize: Dict[int, float]={}, rounds: int=1, amount: int=1):
        super().__init__(name, seeding, solver, cashprize)
        self.nb_rounds = rounds
        self.nb_duel= amount

    def _init_future_rounds(self):
        participants = [p for p in self.participants]
        for _ in range(self.nb_rounds):
            random.shuffle(participants)
            new_round = participants[:self.nb_duel*2]
            
        self.future_rounds.append(new_round)



from .competition import Competition

from .groups import RoundRobin, SwissRound
from .random import RandomRound
from .swissbracket import SwissBracket

from .knockout import SingleEliminationBracket, DoubleEliminationBracket
from .snake import Snake



from typing import List, Callable
from typeguard import typechecked

import abc

from rstt.stypes import SMatch
from rstt.player import Player
import rstt.utils.functions as uf

import numpy as np
import random


class PlayerTVS(Player, metaclass=abc.ABCMeta):
    def __init__(self, name: str, level: float) -> None:
        super().__init__(name=name, level=level)
        self.__current_level = self._BasicPlayer__level
        self.__level_history = [self.__current_level]
        self._Player__games = [None]
        
    # --- getter --- #
    def level_history(self) -> List[float]:
        return self.__level_history
    
    def original_level(self) -> float:
        return self._BasicPlayer__level
    
    def level_in(self, game: SMatch) -> float:
        return self.__level_history[self._Player__games.index(game)]
    
    # --- setter --- #
    def update_level(self, *args, **kwars) -> None:
        self._update_level(*args, **kwars)
        self.__level_history.append(self.__current_level)
        self._Player__add_game(None)

    # --- override --- #
    def level(self) -> float:
        return self.__current_level
    
    def games(self) -> list[SMatch]: #Match
        return [game for game in self.games() if game != None]
    
    def add_game(self, *args, **kwars) -> None:
        super().add_game(*args, **kwars)
        self.__level_history.append(self.__current_level)
    
    def reset(self):
        self._reset_level()
        super().reset()
        self.__level_history.append(self.__current_level)
    
    # --- internal mechanism --- #
    def _reset_level(self) -> None:
        self.__level_history = []
        self.__current_level = self._BasicPlayer__level
    
    @abc.abstractmethod
    def _update_level(self) -> None:
        '''change the self.__current_level value'''
        
class ExponentialPlayer(PlayerTVS):
    @typechecked
    def __init__(self, name: str, start: float, final: float, tau: float):
        super().__init__(name=name, level=start)
        # start to end function of time
        self.__final = final
        self.__time = 0
        self.__relax = uf.exponential_decay
        
        # parameters of the relaxation
        self.__tau = tau
        
    def update_level(self, *args, **kwars) -> float:
        self._time += 1
        self._PlayerTVS__current_level = self.__final - (self.__final - self._PlayerTVS__current_level)*self.__relax(time=self.__time, tau=self.__tau)


class LogisticPlayer(PlayerTVS):
    def __init__(self, name: str, start: float, final: float, center_x: float, r: float, shift: float=0):
        super().__init__(name=name, level=start)
        # start to end function of time
        self.__final = final
        self.__time = 0
        self._relax = uf.verhulst 
        
        # parameters of the relaxation
        self.__tau=uf.a_from_logistic_center(center_x, r)
        self.__r = r
        self.__shift = shift
 
    def update_level(self, *args, **kwars) -> float:
        self.__time += 1
        self._PlayerTVS__current_level += self._relax(K=self.__limit- self._PlayerTVS__current_level,
                                                      a=self.__tau, r=self.__r, t=self.__time, shift=self.__shift)


class CyclePlayer(PlayerTVS):
    """Cycle Player


    Implement the 'Cycle Model' descirbed in
    Aldous D. in 'Elo ratings and the Sports Model: A Negleted Topic in Applied Probability?'
    
    Cycle player have a deterministic level evolution in cycle.
    The variance of the level is given by the attribute __sigma^2,
    while the attribute __tau indicates the number of game needed for the level to decrease
    from its maximum to its avergae value

    Parameters
    ----------
    PlayerTVS : _type_
        _description_
    """
    def __init__(self, name: str, level: float=0, sigma: float=1.0, tau: int=100):
        super().__init__(name, level)
        self.__time = 0
        self.__sigma = sigma # controls the 'variance'
        self.__tau = tau # controls the cycle duration
        
    def update_level(self, *args, **kwars):
        X0 = self._BasicPlayer__level
        self.__time += 1
        self._PlayerTVS__current_level = X0 + uf.deterministic_cycle(mu=X0, sigma=self.__sigma, tau=self.__tau, time=self.__time)


class JumpPlayer(PlayerTVS):
    """Jump Player


    Implement a 'Jump Model' adapted from 
    Aldous D. in 'Elo ratings and the Sports Model'
    
    A JumpPlayer level remains constant for an amount of time given by a geometric distribution
    before 'jumping' to a new level given by a Normal distribution. 

    The class differs from the source document by jumping mulitple times as simulation progress, and not just once.    

    Parameters
    ----------
    PlayerTVS : _type_
        _description_
    """
    def __init__(self, name: str, level: float=0, sigma2: float=1.0, tau: int=400):
        super().__init__(name, level)
        self.__sigma2 = sigma2
        self.__tau = tau
        self.__timer = np.random.geometric(1/self.__tau)
    
    def update_level(self, *args, **kwars):
        self.__tictac()
        self.__jump()
    
    def __tictac(self) -> None:
        self.__timer -= 1
    
    def __jump(self):
        if self.__timer == 0:
            # new timer
            self.__timer = np.random.geometric(1/self.__tau)
            # new level
            self._PlayerTVS__current_level += random.gauss(0, self.__sigma2)




from typeguard import typechecked

from .playerTVS import PlayerTVS

import random

class GaussianPlayer(PlayerTVS):
    @typechecked
    def __init__(self, name: str, mu: float, sigma: float) -> None:
        # pass mu as level to Player, and mu/sigam as params to PlayerTVS
        super().__init__(name=name, level=mu)
        self.__sigma = sigma
        
    def update_level(self, *args, **kwars) -> None:
        self._PlayerTVS__current_level = random.gauss(self._BasicPlayer__level, self.__sigma)


from typeguard import typechecked
from typing import Any, Union, List, Callable, Optional

from rstt import BasicPlayer
from rstt.ranking import Standing
from rstt.stypes import Inference, RatingSystem, Observer



def set_equi(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper_set(self, *args: Any, **kwargs: Any) -> Any:
        set_action = func(self, *args, **kwargs)
        if self._Ranking__maintain_equivalence:
            self._Ranking__ContainerEquivalence()
        return set_action
    return wrapper_set

def get_equi(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper_get(self, *args: Any, **kwars: Any) -> Any:
        if self._Ranking__maintain_equivalence:
            self._Ranking__ContainerEquivalence()
        return func(self, *args, **kwars)
    return wrapper_get
    
def get_disamb(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper_get(self, *args: Any, **kwars: Any) -> Any:
        if self._Ranking__maintain_disambiguity:
            self._Ranking__RankDisambiguity()
        return func(self, *args, **kwars)
    return wrapper_get
    
def set_disamb(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper_set(self, *args: Any, **kwargs: Any) -> Any:
        set_action = func(self, *args, **kwargs)
        if self._Ranking__maintain_disambiguity:
            self._Ranking__RankDisambiguity()
        return set_action
    return wrapper_set


class Ranking():
    '''   
    NOTE:
        - The Ranking class has two dict[Player, ...], (RatingSystem, Standing) - which induce redundancy and ambiguity.
        The 'Container Equivalence' and 'Rank Disambiguity' (SEE GLOSSARY) are two properties that needs to be enforced
        to guarantee a well defined behaviour for the Ranking class. Hopefully it is also sufficient.
        - The RatingSystem act like a defaultdict. Get operations can induce set operations.
        This is tricky as a player could be inserted in the ranking using 'read' methods of the interface.
        
    GLOSSARY:
        - Container Equivalence (Union == Intersection):
        (key in self.datamodel.ratings) <=> (key in self.standing).
        In the code we refer to 'equivalence'
        - Rank Disambiguity (point '=' rating):
        self.datamodel.ordinal(key) == self.standing.value(key) for all keys.
        In the code we refer to 'disambiguity'

    # !!!
        - self.__disambiguity == True && self._Standing__sorted == False
        This means that the players have the correct points but are not correctly ranked.
        It can happen when 'self.standing._keep_sorted_ == False' which is triggered by the user outside of the Ranking methods.
    
    # ??? 
        - Can this state be reached from the intended Ranking usage ?
        How to detect it and what to do ?
    '''
    @typechecked
    def __init__(self, name: str, 
                 datamodel: RatingSystem, 
                 backend: Inference,
                 handler: Observer,
                 players: Optional[List[BasicPlayer]] = None):
        
        # name/identifier - usefull for plot
        self.name = name
        
        # fundamental notion of the Ranking Class
        self.standing = Standing()
        self.backend = backend
        self.datamodel = datamodel
        self.handler = handler

        # state control variable
        self.__equivalence = True
        self.__disambiguity = True

        # protocol control variable
        self.__maintain_equivalence = True
        self.__maintain_disambiguity = True
        
        if players:
            self.add(keys=players)
        
    # --- Containers standard methods --- #
    @set_disamb
    @set_equi
    @typechecked
    def add(self, keys: List[BasicPlayer]):
        # turn maintainance for optimization
        should_maintain = self.__maintain_equivalence
        self.__maintain_equivalence = False
        
        # perform iteratively addition
        for key in keys:
            self.__add(key)
        
        # restaure Ranking status
        self.__maintain_equivalence = should_maintain
             
    def __add(self, key: BasicPlayer):
        if key in self:
            msg = f'Can not add a key already present in the Ranking, {key}'
            raise ValueError(msg)
        # default dict get operator for missing key
        self.datamodel.get(key)
        
        # self.datamodel do not match self.standing
        self.__equivalence = False
    
    # --- magic methods --- #
    def __getitem__(self, *args, **kwargs) -> Union[BasicPlayer, List[BasicPlayer], int, List[int]]:
        ''' get item based on a rank or a key
        
        NOBUG:
            - sorting is handled by the Standing itself
            - so is typechecked
        '''
        return self.standing.__getitem__(*args, **kwargs)
    
    def __delitem__(self, key: BasicPlayer):
        ''' delete element from the Ranking
        
        REQ:
            - element needs to be remove both from the standing and the RatingSystem
            
        NOBUG:
            - del standing[key] is typechecked. CALL MUST BE BEFORE RatingSystem
        
        # ???:
            - could del succeed on standing but fail on RatingSystem.
            This would otentialy lead to an invalid ranking state because __delitem__ is not decorated.
            A ranking invalid state can exactly be the reason why this scenario happens.
            What is the best approach to this situation ?
        '''
        del self.standing[key]
        del self.datamodel[key]
            
    def __contains__(self, key: BasicPlayer):
        '''
        NOTE:
            - it does match standing behavior as specified but
            depending on the choices, 'p in self' can be slower than
            'p in self.standing' / 'p in self.datamodel'
        '''
        return key in self.standing
    
    def __len__(self):
        return len(self.standing)
    
    def __iter__(self):
        return self.standing.__iter__()
                
    # --- getter --- #
    def rank(self, player: BasicPlayer) -> int:
        return self[player]
        
    def ranks(self, players: List[BasicPlayer]) -> List[int]:
        return [self.rank(player) for player in players]
        
    def rating(self, player: BasicPlayer) -> Any:
        """Get method for rating

        Rating object is the internal model associated to a key.
        Ratings are used to automaticly compute values for the sorting feature of a Standing.

        Parameters
        ----------
        player : Player
            A key in the Ranking

        Returns
        -------
        _type_
            The associated model to the provided key. The type is defined by Ranking.RatingSystem.rtype
            
        Raises
        ------
        KeyError
        """
        if player in self: # NOBUG RatingSystem is a defaultdict
            return self.datamodel.get(player)
        else:
            msg = f"{player} is not present in {self.standing}"
            raise KeyError

    def ratings(self) -> List[Any]:
        """Get method for all ratings

        Returns
        -------
        list[_type_]
            A list of all rating object present in the Ranking, in order of the Standing.
        """
        return [self.rating(player) for player in self]
    
    def players(self) -> List[BasicPlayer]:
        """Get method of all keys

        Alias for Ranking.standing.keys()

        Returns
        -------
        List[Player]
            A list of all player in descending order of their associated values.
        """
        return self.standing.keys()

    def point(self, player: BasicPlayer) -> float:
        """Get the point associated to a key

        Alias for Ranking.standing.value(player)

        Returns
        -------
        float
            the associated value.
        """
        return self.standing.value(player)

    def points(self) -> List[float]:
        """Get method of all values

        Alias for Ranking.standing.values()

        Returns
        -------
        List[float]
            A list of all associated values in descending order.
        """
        return self.standing.values()

    def status(self):
        return  {'equivalence': self.__equivalence,
        'disambiguity': self.__disambiguity,
        'maintain_equivalence': self.__maintain_equivalence,
        'm_disambanbiguity': self.__maintain_disambiguity}
        
    # ??? items() -> List[(rank, player, ratings)]
    # ??? item(key) -> (rank, player, ratings)
    # --- setter --- #
    @set_disamb
    @set_equi
    def set_rating(self, key: BasicPlayer, rating: Any):
        """A method to assign a rating to a Player

        
        The Ranking delegate this task to a 'RatingSystem' instance stored as attribute 'rankink.datamodel'.
        The RatingSystem define what rating type is accepted and wether a set operation is authorized for the provided key.
        
        Parameters
        ----------
        key : Player
            A Player
        rating : Any
            A rating object associated to the key
        """
        self.datamodel.set(key, rating)
        self.__equivalence = False
        
    # ??? remove()
    # ??? pop()
            
    # --- general purpose methods --- #
    @get_disamb
    @get_equi   
    def plot(self):
        self.standing.plot(standing_name=self.name)
    
    @set_disamb
    @set_equi
    def update(self, *args, **kwargs):
        """Update the Ranking


        Parameters
        ----------
        observations : _type_
            _description_
        """
        
        self.forward(*args, **kwargs)
      
        # NOTE: How do we know if the ranking state changed ?
        # HACK: always assume it did
        self.__disambiguity = False
        self.__equivalence = False
    
    def forward(self, *args, **kwargs):
        '''
        '''
        self.handler.handle_observations(infer=self.backend,
                            datamodel=self.datamodel,
                            *args, **kwargs)

    @get_disamb
    @get_equi
    def rerank(self, permutation: List[int], name: str = None, direct: bool = True):
        # TODO: write doc
        # check permutation validity
        if not (set(permutation) == set(list(range(len(self))))):
            # TODO: write good msg error
            msg = ''
            raise ValueError(msg)
        
        # TODO: deal with direct=False
        
        pairs = []
        for current_rank, future_rank in enumerate(permutation):
            player = self[current_rank]
            ratings = self.datamodel.get(self[future_rank])
            pairs.append((player, ratings))
        for p, r in pairs:
            self.datamodel.set(p, r)
    
    @get_disamb
    @get_equi
    def fit(self, players: List[BasicPlayer], name: str = ''):
        seeding= Standing()
        points = []
        for player in players:
            if player in self:
                points.append(self.point(player))
            else:
                points.append(self.standing._default_)
        seeding.add(players, points)
        return seeding
    
    # --- internal mechanism --- #
    def __ContainerEquivalence(self):
        ''' property checker
        
        '''
        # get keys
        standing_keys = set(self.standing.keys())
        RatingSystem_keys = set(self.datamodel.keys())
        
        if standing_keys == RatingSystem_keys:
            self.__equivalence = True
        elif standing_keys <= RatingSystem_keys: # a <= b means 'a.issubset(b)
            not_ranked_players = list(RatingSystem_keys - standing_keys) # a - b means a.difference(b)
            new_points = []
            for player in not_ranked_players:
                # NOBUG: player is in the RatingSystem. 'get()' is safe to perform.
                new_points.append(self.datamodel.ordinal(self.datamodel.get(player)))
            # NOBUG: no ambiguity is introduce this way
            self.standing.add(keys=not_ranked_players, values=new_points)
            self.__equivalence = True
        else:
            # TODO: write a good error message
            msg = f''
            raise RuntimeError(msg)
        
    def __RankDisambiguity(self):
        for player in self.standing:
            # TODO: performance check (a) assign only if needed (b) assign always
            # ??? (a / b) as user options ?
            rating = self.datamodel.get(player)
            point = self.datamodel.ordinal(rating)
            if self.standing.value(player) != point:
                self.standing[player] = point
        
        self.__disambiguity = True




from typing import List, Dict, Callable, Any
from typeguard import typechecked


from . import Competition
from rstt.ranking.ranking import Ranking
from rstt import Player, BetterWin
from rstt.stypes import Solver

from rstt.utils import utils as uu, matching as um, competition as uc

import math

def balanced_tree(rounds):
    matches = [1, 2]
    for round in range(2, rounds+1):
        check_sum = pow(2, round)+1
        new_matches = []
        for i in matches:
            new_matches.append([i, check_sum-i])
        matches = uu.flatten(new_matches)

    return matches


class SingleEliminationBracket(Competition):
    @typechecked
    def __init__(self, name: str,
                seeding: Ranking,
                solver: Solver = BetterWin(),
                cashprize: Dict[int, float] = {}):
        super().__init__(name, seeding, solver, cashprize)

    # --- override --- #
    def _initialise(self):
        msg = (f'{type(self)} '
                'needs a power of two as number of participants '
                '(2,4,8,16,...)'
                f', given {len(self.participants)}')
        assert uu.power_of_two(len(self.participants)), msg

        nb_rounds = int(math.log(len(self.participants), 2))
        self.players_left = self.seeding[[i-1 for i in balanced_tree(nb_rounds)]]

    def generate_games(self):
        return uc.playersToDuel(self.players_left)
    
    def _end_of_stage(self) -> bool:
        return True if len(self.players_left) == 1 else False
    
    def _update(self):
        next = [game.winner() for game in self.played_matches[-1]]
        self.players_left = next
    
    def _standing(self) -> Dict[Player, int]:
        standing = {}
        top = len(self.participants)
        for round in self.played_matches:
            for game in round:
                standing[game.loser()] = top
            top = len(self.participants) - len(standing)
        
        # winner
        standing[self.played_matches[-1][0].winner()] = 1
        return standing


class DoubleEliminationBracket(Competition):
    @typechecked
    def __init__(self, name: str,
                seeding: Ranking,
                solver: Solver = BetterWin(),
                lower_policy: Callable[[List[Any]], List[Any]]=lambda x: x ,
                injector_policy: Callable[[List[Any], List[Any]], List[Any]]=um.riffle_shuffle,
                cashprize: Dict[int, float] = {}):
        super().__init__(name, seeding, solver, cashprize)
        
        self.upper = SingleEliminationBracket(name+'_UpperBracket',
                                              seeding, solver)
        self.lower = [] # List[Player]
        self.lower_policy = lower_policy
        self.injector_policy = injector_policy

    # --- override --- #
    def _initialise(self):
        # NOBUG: do notrun(). Not 'event' in itself -> no upper.trophies() called
        self.upper.registration(self.participants)
        self.upper.start()
        self.upper.play()
        
        # lower bracket
        self.lower = [[game.loser() for game in r] for r in self.upper.games(by_rounds=True)]
        self.lower += [[self.upper.games(by_rounds=True)[-1][0].winner()]]

    def _update(self):
        self.lower.insert(0, [game.winner() for game in self.played_matches[-1]])
    
    def _standing(self) -> Dict[uc.Player, int]:
        standing = {}
        top = len(self.participants)
        for round in self.played_matches:
            for game in round:
                standing[game.loser()] = top
            top = len(self.participants) - len(standing)
        # winner
        standing[self.played_matches[-1][0].winner()] = 1
        return standing

    def _end_of_stage(self) -> bool:
        return len(self.lower) == 1

    def generate_games(self):
        if len(self.lower[0]) != len(self.lower[1]):
            # lower bracket games
            games = uc.playersToDuel(self.lower_policy(self.lower.pop(0)))
        else:
            # injector games
            lower = self.lower.pop(0)
            injector = self.lower.pop(0)
            games = uc.playersToDuel(self.injector_policy(lower, injector))
        return games
    
    @typechecked
    def games(self, by_rounds=False, upper=False, lower=False):
        if upper and lower:
            msg = f"At most one of upper and lower can be True. Received values upper: {upper}, lower: {lower}"
            raise ValueError(msg)
        if upper:
            return self.upper.games(by_rounds)
        elif lower:
            return super().games(by_rounds)
        else:
            games = self.upper.played_matches + self.played_matches
            if by_rounds:
                return games
            else:
                return uu.flatten(games)



from typing import List, Dict
from typeguard import typechecked

from rstt import Duel, BetterWin
from . import Competition
from rstt.ranking.ranking import Ranking
from rstt.stypes import Solver
from rstt.solver.solvers import WIN, LOSE, DRAW

from rstt.utils import utils as uu, matching as um, competition as uc

import numpy as np
import math

class RoundRobin(Competition):
    def __init__(self, name: str, seeding: Ranking, solver: Solver = BetterWin(),
                cashprize: Dict[int, float] = {}):
        
        super().__init__(name, seeding, solver, cashprize)
        
        self.table = None
        self.future_rounds = []
    
    # --- Override --- #
    def _initialise(self):
        self._init_table()
        self._init_future_rounds()
        
    def _end_of_stage(self):
        return not self.future_rounds

    def _update(self):
        for game in self.played_matches[-1]:
            p1, p2 = game.player1(), game.player2()
            s1, s2 = self.seeding[[p1, p2]]
            self.table[s1][s2] += game.score(p1)
            self.table[s2][s1] += game.score(p2)
 
    def _standing(self):
        standing = {}
        groups = []
        
        scores = np.sum(self.table, axis=0)
        values = np.unique(scores)
        for value in values:
            index = np.where(scores == value)[0].tolist()
            groups.append(self.seeding[index])
        
        top = 0
        for group in groups:
            top += len(group)
            standing.update({player: top for player in group})
            
        return standing
        
    def generate_games(self):
        return self.next_round()
        
    # --- init stuff --- #
    def _init_table(self):
        self.table = np.zeros(shape=(len(self.participants), len(self.participants)))
    
    def _init_future_rounds(self):
        self.future_rounds = um.ruban([p for p in self.seeding if p in self.participants])

    # --- round mechanisme --- #
    def next_round(self):
        # FIXME: seems unecessary -> this code in generate_games(self)
        games = uc.playersToDuel(self.future_rounds.pop(0))
        return games


class SwissRound(RoundRobin):
    def _init_future_rounds(self):
        self.future_rounds = [[player for player in self.seeding]]

    # --- Override --- #
    def _end_of_stage(self):
        return len(self.played_matches) == int(math.log(len(self.participants), 2))
    
    def _update(self):
        # !!! not how _end_of_stage() is meant to be used.
        if not self._end_of_stage():
            self.make_groups()

    def next_round(self):
        games = [uc.find_valid_draw(draws=self.draws(group),games=self.games()) for group in self.future_rounds]
        return uu.flatten(games)
    
    # --- round mechanisme --- #
    def make_groups(self):
        self.future_rounds = []
        scores = np.sum(self.table, axis=1)
        values = np.unique(scores)
        for value in values:
            # build a round
            index = np.where(scores == value)[0].tolist()
            players = self.seeding[index]
            self.future_rounds.append(players)
    
    def draws(self, players):
        # FIXME: explore other methods / make it tunable. It could result in bug
        return [uc.playersToDuel(round) for round in um.ruban(players)]
    


from typing import  List, Dict, Tuple, Union,Optional, Any
from typeguard import typechecked

import rstt.config as cfg
from rstt.player import Player
from rstt.stypes import SPlayer, Event

import numpy as np
import math
import copy

import warnings


'''
    FIXME:
    - There is no consensus on the return type of the rate() functions
    - It is fine regarding the Protocol, but realy anoying when combine with Observer
'''


# ------------------------ #
# --- Player as Rating --- #
# ------------------------ #    
class PlayerLevel:
    @typechecked
    def rate(self, player: SPlayer, *args, **kwars) -> Dict[Player, float]:
        return {player: player.level()}
    
class PlayerWinPRC:
    def __init__(self, default: float=-1.0, scope: int=np.iinfo(np.int32).max):
        self.default = default
        self.scope = scope
    
    @typechecked
    def rate(self, player: Player, *args, **kwargs):
        return {player: self.win_rate(player)}
    
    def win_rate(self, player: Player):
        games = player.games()
        if games:
            if self.scope:
                games = games[-self.scope:]
            nb_wins = sum([1 for game in games if player is game.winner()])
            total = len(games)
            winrate = nb_wins / total * 100
        else:
            winrate = self.default
        return winrate


# ------------------------ #
# ----- Event Based ------ #
# ------------------------ #
class EventStanding:
    def __init__(self, buffer: int, best: int, default: Optional[Dict[int, float]]=None):
        self.buffer = buffer
        self.best = best
        
        self.events = []
        self.points = {}
        
        self.__default_points = default if default else cfg.EVENTSTANDING_DEFAULT_POINTS
    
    @typechecked
    def add_event(self, event: Union[Event, str], points: Optional[Dict[int, float]]=None):
        points = points if points else self.__default_points
        if isinstance(event, str): # by str
            self.events.append(event)
            self.points[event] = points
        else: # by Event
            self.events.append(event.name())
            self.points[event.name()] = points
    
    @typechecked
    def remove_event(self, event: Union[Event, str]):
        if isinstance(event, str): # by str
            self.events.remove(event)
            del self.points[event]
        else: # by Event
            self.events.remove(event.name())
            del self.points[event.name()]
    
    @typechecked
    def rate(self, player: Player):
        # events that matter
        events = self.events[-self.buffer:]
        
        # collected points in events
        results = []
        for achievement in player.achievements():
            if achievement.event_name in events:
                results.append(self.points[achievement.event_name][achievement.place])

        # get only the best results
        results.sort()
        best_results = results[-min(len(results),self.best):]
        
        # sum the best results considered
        points = sum(best_results)
        return {player: points}

    
# ------------------------ #
# --- Game Score Based --- #
# ------------------------ #
class Elo:
    def __init__(self, k: float = 20.0, lc: float = 400.0):
        self.LC = lc
        self.K = k
        # TODO self.distribution = dist & change expectedScore

    @typechecked
    def rate(self, groups: List[List[float]], scores: List[float], *args, **kwars) -> List[List[float]]:        
        # NOTE: groups: [[winner_elo][loser_elo]], scores [[1.0][0.0]]
        
        # unpack args
        r1, r2 = groups[0][0], groups[1][0]
        s1, s2 = scores
        # cumpute new ratings
        new_r1 = self.update_rating(r1, r2, s1)
        new_r2 = self.update_rating(r2, r1, s2)
        return [[new_r1], [new_r2]]

    def expectedScore(self, rating1, rating2):
        return 1.0 / (1.0 + math.pow(10, (rating2-rating1)/self.LC))
    
    def update_rating(self, rating1: float, rating2:float, score:float):
        expected_result = self.expectedScore(rating1, rating2)
        return rating1 + (self.K * (score-expected_result))


class Glicko:
    def __init__(self, minRD: float = 30.0,
                maxRD: float = 350.0,
                c: float = 63.2,
                q: float = math.log(10, math.e)/400,
                lc: int = 400):
        
        # model constant
        self.__maxRD = maxRD # maximal value of RD
        self.__minRD = minRD # minimal value of RD
        self.lc = lc # constant in function E
        self.C = c # constant used for 'inactivity decay'
        self.Q = q # no idea how to interpret this value
        
    def G(self, rd: float):
        return 1 / math.sqrt( 1 + 3*self.Q*self.Q*(rd*rd)/(math.pi*math.pi))
    
    def expectedScore(self, rating1, rating2, mode='update'):
        RDi = 0 if mode == 'update' else rating1.sigma
        RDj = rating2.sigma
        ri, rj = rating1.mu, rating2.mu
        return 1 / (1 + math.pow(10, -self.G(math.sqrt(RDi*RDi + RDj*RDj)) * (ri-rj)/400))

    def d2(self, rating1, games: List[Tuple[Any, float]]):
        ''' NOTE:
        rating have mu sigma attributes, 
        games are dict of ratings:score,
        '''
        all_EJ = []
        all_GJ = []
        for rating2, score in games:
            # get needed variables
            Ej = self.expectedScore(rating1, rating2, mode='update')
            RDj = rating2.sigma
            Gj = self.G(RDj)
            
            # store vairables
            all_EJ.append(Ej)
            all_GJ.append(Gj)
        
        # big sum
        bigSum = 0
        for Gj, Ej, in zip(all_GJ, all_EJ):
            bigSum += Gj*Gj*Ej*(1-Ej)
        
        '''
        NOTE:
        Try/Expect is not part of the Glicko official algorith  presentation.
        But I have encountered Unexpected ZeroDivisionError
        
        The try/except ZeroDivErr/warn an be easly simplified by:
        return 1 / min( self.Q*self.Q*bigSum, lower_bound)
        
        However I could note find any specfic details about the choice of the boundary.
        Analytically, the term can not be equal to 0.0, it is always >0.
        Nnumercialy, it happens in extreme situation i.e does not arise in standard 'intended' Glicko usage.
        
        The package is for scientifical experimentation,
        It allows extreme case exploration and can not hide arbitrary choices.

        # !!! Do not fix unless it is possible to link a scientifical source justifying the implementation
        '''
        try:
            # d2 formula 
            return 1 / (self.Q*self.Q*bigSum)
        except ZeroDivisionError:
            # !!! BUG: ZeroDivisionError observed with extreme rating differences
            # !!! this will now print variable of interest
            # !!! but code will run assuming maximal and mininal expected value possible between 0 and 1

            # HACK: just assume a very low 'bigSum'
            bigSum = 0.00000000001
            correction = 1 / (self.Q*self.Q*bigSum)
            
            msg = f"Glicko d2 ERROR: {rating1}, {games}\n {bigSum}, {all_EJ}, {all_GJ}\n d2 return value as been adjusted to 1/0.00000000001 "
            warnings.warn(msg, RuntimeWarning)
            return correction

    def prePeriod_RD(self, rating):
        '''
        implement formula presented at step1 caee (b) p.3
        '''
        new_RD = math.sqrt(rating.sigma*rating.sigma + self.C*self.C)
        # check boundaries on sigma - ??? move max() elsewhere
        return max(min(new_RD, self.__maxRD), self.__minRD)
    
    def newRating(self, rating1, games: List[Tuple[Any, float]]):
        ''' NOTE:
        rating have mu sigma attributes, 
        games are dict of ratings:score,
        '''
            
        # compute term 'a'
        d2 = self.d2(rating1, games)
        a = self.Q / ((1/(rating1.sigma*rating1.sigma)) + (1/d2))
        
        # lcompute term 'b'
        b = 0
        for rating2, score in games:
            b += self.G(rating2.sigma)*(score - self.expectedScore(rating1, rating2, mode='update'))
        
        # create new rating object to avoid 'side effect'
        rating = copy.copy(rating1)
        # post Period R
        rating.mu += a*b
        # post Period RD
        rating.sigma = math.sqrt(1/( (1/rating1.sigma**2) + (1/d2) ))
        
        return rating

    def rate(self, rating, ratings: List[Any], scores: List[float], *args, **kwars):
        # formating
        games = [(r, s) for r, s in zip(ratings, scores)]
        return self.newRating(rating, games)





from . import rating

from .standing import Standing
from .ranking import Ranking

from .observer import GameByGame, BatchGame, KeyChecker
from .inferer import Elo, Glicko, PlayerLevel, PlayerWinPRC, EventStanding
from .datamodel import KeyModel, GaussianModel

from .standard import BTRanking, WinRate, SuccessRanking
from .standard import BasicElo, BasicGlicko, BasicOS



from typing import List, Dict

from rstt import Duel, Player
from . import Competition


class Snake(Competition):
    def __init__(self, *args, **kwars):
        super().__init__(*args, **kwars)
        self.snake = []

    def _initialise(self) -> None:
        self.snake = [player for player in self.seeding]
        self.snake.reverse()
    
    def _update(self) -> None:
        self.snake.insert(0, self.played_matches[-1][0].winner())

    def _end_of_stage(self) -> bool:
        return len(self.snake) == 1
    
    def _standing(self) -> Dict[Player, int]:
        standing = {games[0].loser(): len(self.participants)-r
                         for r, games in enumerate(self.played_matches)}
        standing[self.played_matches[-1][0].winner()] = 1
        return standing
        
    def generate_games(self) -> List[Duel]:
        return [Duel(self.snake.pop(0), self.snake.pop(0))]


        
    '''
    def edit(self, games: List[Duel]):
        for game in games:
            # winner can will play another game
            self.snake.insert(0, game.winner())
            # loser journey ends
            self.standing[game.loser()] = len(self.snake) + 1

        # check stop condition
        if len(self.snake) == 1:
            self.standing[self.snake[0]] = 1
            finished = True
        else:
            finished = False

        return finished'''


from typing import Union, List, Dict, Tuple, Optional, Any, Iterable
from typing import Protocol, runtime_checkable
from typeguard import typechecked

from rstt import Player, Duel, BetterWin
from rstt.ranking.ranking import Ranking
from rstt.stypes import Solver, Shuffler, Seeder, Generator, Evaluator
from . import Competition

from rstt.utils import utils as uu, matching as um, competition as uc


class DummyParam(Shuffler, Seeder, Evaluator):
    def rearange(status: List[int]) -> List[int]:
        return status

    def seed(players: List[Player], *args,  **kwargs) -> List[Player]:
        return players

    def generate(status: Union[List[int], List[Player]]) -> Union[List[int], List[Player]]:
        return um.ruban(status)
    
    def eval(options: List[List[Player]], *args,  **kwargs) -> List[List[Player]]:
        return options

class SwissBracket(Competition):
    @typechecked
    def __init__(self, name: str, seeding: Ranking, solver: Solver = BetterWin(),
                 matchings: Optional[Dict[Tuple[int, int], Shuffler]]=None,
                 seeders: Optional[Dict[Tuple[int, int], Seeder]]=None,
                 generators: Optional[Dict[Tuple[int, int], Generator]]=None,
                 evaluators: Optional[Dict[Tuple[int, int], Evaluator]]=None,
                 def_matching=DummyParam,
                 def_seeder=DummyParam,
                 def_generator=DummyParam,
                 def_evaluator=DummyParam,
                 cashprize: Dict[int, float] = {}):
        super().__init__(name, seeding, solver, cashprize)

        # general settings of the system
        self.current_round = 0
        self.max_round = None
        self.max_wins = None
        self.max_loses = None
    
        # round parameters - map score with data: (int, int) -> data
        self.rounds = {} # List[Player]
        self.matchings = matchings if matchings else {} # Callable
        self.seeders = seeders if seeders else {} # Callable
        self.generators = generators if generators else {} # Callable
        self.evaluators = evaluators if evaluators else {} # Callable
        
        self.def_matching = def_matching
        self.def_seeder = def_seeder
        self.def_generator = def_generator
        self.def_evaluator = def_evaluator
                
    # --- Override ---#
    def _initialise(self):
        # !!! Currently hardcoded setting for 16 participants
        msg = f"Currently SwissBracket is implemented only for 16 players, received {len(self.participants)}"
        assert len(self.participants) == 16, msg
        
        self.__init_settings()        
        self.__init_params()
        
    def _update(self) -> None:
        # ??? clear self.rounds or keep history 
        # move winner and loser
        for game in self.played_matches[-1]:
            p1, p2 = game.player1(), game.player2()
            self.rounds[self.score(p1)].append(p1)
            self.rounds[self.score(p2)].append(p2)
            
        # move to next round
        self.current_round += 1

    def _end_of_stage(self) -> bool:
        return self.current_round == self.max_round
    
    def _standing(self) -> Dict[Player, int]:
        standing = {}
        top8 = [(self.max_wins, i) for i in range(self.max_loses)]
        bottom8 = [(i, self.max_loses) for i in range(self.max_wins-1, -1, -1)]
        final_scores = top8 + bottom8
        top = 0
        for score in final_scores:
            players = self.rounds[score]
            top += len(players)
            for p in players: 
                standing[p] = top
        return standing

    def generate_games(self):
        games = [uc.find_valid_draw(draws=self.draws(group=self.rounds[score]),
                                    games=self.games()) 
                                    for score in self.round_scores()]
        return uu.flatten(games)
    
    # --- round mechanism --- #
    @typechecked
    def score(self, player: Player) -> Tuple[int, int]:
        wins = len([game for game in self.games() if game.winner() == player])
        loses = len([game for game in self.games() if game.loser() == player])
        return (wins, loses)

    def round_scores(self) -> List[Tuple[int, int]]:
        # Round1 [(0, 0)]; Round2 [(1,0), (0,1)]; Round3 [(2,0), (1,1), (0,2)] ...
        return [(self.current_round-i, i) for i in range(self.current_round+1)
                if i < self.max_loses and self.current_round-i < self.max_wins]

    @typechecked
    def draws(self, group: List[Player]) -> List[List[Duel]]:
        scores = [self.score(player) for player in group]
        assert len(set(scores)) == 1 # !!! debugging - SwissBracket will change to support different amount of players
        score = scores[0]        
        
        # sellect correct round parameters
        matching = self.matchings[score].rearange
        seeder = self.seeders[score].seed
        generator = self.generators[score].generate
        evaluator = self.evaluators[score].eval

        # how do seeds face each other by default
        paired_list = matching(list(range(len(group))))
        
        # who has which seeds - replace 'int' by 'Player'
        paired_player = seeder(players=[group[i] for i in paired_list],
                               initial_seeds=self.seeding,
                               results=self.results())
        
        # what are all the options we want to check
        options = generator(paired_player)
       
        # in which order should we consider them
        options = evaluator(options,
                            initial_seeds=self.seeding,
                            results=self.results())
        
        return[uc.playersToDuel(option) for option in options]

    def results(self):
        # fit data to expected format for tiebreaker function
        res = {player: {'opponent': [], 'score': [],} for player in self.seeding}
       
        for game in self.games():
            # update opponent list
            res[game.player1()]['opponent'].append(game.player2())
            res[game.player2()]['opponent'].append(game.player1())
            
            # update score list !!! Lacks generalisation
            res[game.player1()]['score'].append('win' if game.winner() == game.player1() else 'lose')
            res[game.player2()]['score'].append('win' if game.winner() == game.player2() else 'lose')
        return res
    
    # --- under the hood mechanism --- #    
    def __possible_scores(self):
        return [(i,j) for j in range(self.max_loses+1) for i in range(self.max_wins+1) if i+j <= self.max_round]

    def __init_settings(self):
        self.max_round = 5
        self.max_wins = 3
        self.max_loses = 3
        
    def __init_params(self):
        scores = self.__possible_scores()
        
        self.rounds[(0, 0)] = [player for player in self.seeding]
        self.rounds = {score: self.rounds[score] if score in self.rounds.keys()
                       else [] for score in scores}
        
        self.matchings = {score: self.matchings[score] if score in self.matchings.keys()
                                  else self.def_matching for score in scores}
        
        self.seeders = {score: self.seeders[score] if score in self.seeders.keys()
                                  else self.def_seeder for score in scores}
        
        self.generators = {score: self.generators[score] if score in self.generators.keys()
                                  else self.def_generator for score in scores}
        
        self.evaluators = {score: self.evaluators[score] if score in self.evaluators.keys()
                                  else self.def_evaluator for score in scores}